#include "AfxWin.h"
#include "local.h"
#include <stdio.h>
#include <stdlib.h>
#include <X11/Xlib.h>
#include "TConvAttr.h"
#include "XIMInputMethodP.h"
#include "kanji.h"

static	XIMInputStyle	srSupportedStyles []	= {
	{	XIMPreeditPosition  | XIMStatusNothing,		CONVERSION_INPUTSTYLE_OVERTHESPOT, },
	{	XIMPreeditPosition  | XIMStatusArea,		CONVERSION_INPUTSTYLE_OVERTHESPOT, },
	{	XIMPreeditArea      | XIMStatusArea,		CONVERSION_INPUTSTYLE_OFFTHESPOT, },
	{	XIMPreeditNothing   | XIMStatusNothing,		CONVERSION_INPUTSTYLE_ROOTWINDOW, },
	{	XIMPreeditCallbacks | XIMStatusCallbacks,	CONVERSION_INPUTSTYLE_ONTHESPOT, },
	{	XIMPreeditCallbacks | XIMStatusNothing,		CONVERSION_INPUTSTYLE_ONTHESPOT, },
} ;

XIMInputMethod*
XIMInputMethod_Create (
	register Widget		gw,
	register int		nID)
{
	register XIMInputMethod*	pInputMethod ;

	pInputMethod	= MALLOC (sizeof (XIMInputMethod)) ;

	pInputMethod->m_wgClient		= gw ;
	pInputMethod->m_woID			= nID ;
	pInputMethod->m_woICID			= 1 ;
	pInputMethod->m_woExtensionMask	= 0 ;
	pInputMethod->m_nCodingSystem	= KCODING_SYSTEM_COMPOUND_TEXT ;
	pInputMethod->m_lstInputContext	= NULL ;
	return	pInputMethod ;
}

void
XIMInputMethod_Destroy (
	register XIMInputMethod*	pInputMethod)
{
	register XIMInputContext*	pNode ;
	register XIMInputContext*	pNextNode ;

	/*	InputContext ˴롣*/
	assert (pInputMethod != NULL) ;
	pNode	= pInputMethod->m_lstInputContext ;
	while (pNode != NULL) {
		pNextNode	= XIMInputContext_GetNext (pNode) ;
		XIMInputContext_Destroy (pNode) ;
		pNode		= pNextNode ;
	}
	FREE (pInputMethod) ;
	return ;
}

int
XIMInputMethod_GetID (
	register XIMInputMethod*	pInputMethod)
{
	assert (pInputMethod != NULL) ;
	return	pInputMethod->m_woID ;
}

Widget
XIMInputMethod_GetClient (
	register XIMInputMethod*	pInputMethod)
{
	assert (pInputMethod != NULL) ;
	return	pInputMethod->m_wgClient ;
}

XIMInputContext*
XIMInputMethod_CreateInputContext (
	register XIMInputMethod*	pInputMethod)
{
	XIMInputContext*	pInputContext ;

	assert (pInputMethod != NULL) ;
	pInputContext	= XIMInputContext_Create (pInputMethod, pInputMethod->m_woICID) ;
	if (pInputContext == NULL)
		return	NULL ;
	pInputMethod->m_woICID	++ ;

	XIMInputContext_SetNext (pInputContext, pInputMethod->m_lstInputContext) ;
	pInputMethod->m_lstInputContext	= pInputContext ;
	return	pInputContext ;
}

void
XIMInputMethod_DestroyInputContext (
	register XIMInputMethod*	pInputMethod,
	register XIMInputContext*	pInputContext)
{
	register XIMInputContext*	pNode ;
	register XIMInputContext*	pPrevNode ;
	
	assert (pInputMethod != NULL) ;
	pNode		= pInputMethod->m_lstInputContext ;
	pPrevNode	= NULL ;
	while (pNode != NULL) {
		if (pNode == pInputContext) {
			register XIMInputContext*	pNextNode ;

			if (pPrevNode == NULL) {
				pNextNode	= XIMInputContext_GetNext (pNode) ;
				pInputMethod->m_lstInputContext	= pNextNode ;
			} else {
				pNextNode	= XIMInputContext_GetNext (pNode) ;
				XIMInputContext_SetNext (pPrevNode, pNextNode) ;
			}
			XIMInputContext_Destroy (pInputContext) ;
			return ;
		}
		pPrevNode	= pNode ;
		pNode		= XIMInputContext_GetNext (pPrevNode) ;
	}
	return ;
}

XIMInputContext*
XIMInputMethod_FindInputContext (
	register XIMInputMethod*	pInputMethod,
	register int				nID)
{
	register XIMInputContext*	pNode ;

	assert (pInputMethod != NULL) ;
	pNode	= pInputMethod->m_lstInputContext ;
	while (pNode != NULL) {
		if (XIMInputContext_GetID (pNode) == nID)
			return	pNode ;
		pNode	= XIMInputContext_GetNext (pNode) ;
	}
	return	NULL ;
}

int
XIMInputMethod_GetQueryInputStyle (
	register XIMInputMethod*		pInputMethod,
	register const XIMInputStyle**	ppSupportedStyle)
{
	assert (pInputMethod != NULL) ;
	if (ppSupportedStyle != NULL)
		*ppSupportedStyle	= srSupportedStyles ;
	return	NELEMENTS (srSupportedStyles) ;
}

CARD16
XIMInputMethod_GetExtMask (
	register XIMInputMethod*			pInputMethod)
{
	assert (pInputMethod != NULL) ;
	return	pInputMethod->m_woExtensionMask ;
}

Boolean
XIMInputMethod_SetExtMask (
	register XIMInputMethod*			pInputMethod,
	register CARD16						woMask)
{
	assert (pInputMethod != NULL) ;
	pInputMethod->m_woExtensionMask	= woMask ;
	return	True ;
}

void
XIMInputMethod_SetNext (
	register XIMInputMethod*			pInputMethod,
	register XIMInputMethod*			pNextInputMethod)
{
	assert (pInputMethod != NULL) ;
	pInputMethod->m_pNext	= pNextInputMethod ;
	return ;
}

XIMInputMethod*
XIMInputMethod_GetNext (
	register XIMInputMethod*			pInputMethod)
{
	assert (pInputMethod != NULL) ;
	return	pInputMethod->m_pNext ;
}

Boolean
XIMInputMethod_SetCodingSystem (
	register XIMInputMethod*	pThis,
	register int				nCodingSystem)
{
	assert (pThis != NULL) ;
	if (nCodingSystem < 0 || nCodingSystem >= MAX_KCODING_SYSTEM)
		return	False ;
	pThis->m_nCodingSystem	= nCodingSystem ;
	return	True ;
}

int
XIMInputMethod_GetCodingSystem (
	register XIMInputMethod*	pThis)
{
	assert (pThis != NULL) ;

	return	pThis->m_nCodingSystem ;
}





