/*
 * re_*exec and friends - match REs
 *
 * Copyright (c) 1998, 1999 Henry Spencer.  All rights reserved.
 * 
 * Development of this software was funded, in part, by Cray Research Inc.,
 * UUNET Communications Services Inc., Sun Microsystems Inc., and Scriptics
 * Corporation, none of whom are responsible for the results.  The author
 * thanks all of them. 
 * 
 * Redistribution and use in source and binary forms -- with or without
 * modification -- are permitted for any purpose, provided that
 * redistributions in source form retain this entire copyright notice and
 * indicate the origin and nature of any modifications.
 * 
 * I'd appreciate being given credit for this package in the documentation
 * of software which uses it, but that is not a requirement.
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * HENRY SPENCER BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include "regguts.h"
#include "bufstring.h"

/* lazy-DFA representation */
struct arcp {			/* "pointer" to an outarc */
	struct sset *ss;
	color co;
};

struct sset {			/* state set */
	unsigned *states;	/* pointer to bitvector */
	unsigned hash;		/* hash of bitvector */
#		define	HASH(bv, nw)	(((nw) == 1) ? *(bv) : hash(bv, nw))
#	define	HIT(h,bv,ss,nw)	((ss)->hash == (h) && ((nw) == 1 || \
		memcmp(VS(bv), VS((ss)->states), (nw)*sizeof(unsigned)) == 0))
	int flags;
#		define	STARTER		01	/* the initial state set */
#		define	POSTSTATE	02	/* includes the goal state */
#		define	LOCKED		04	/* locked in cache */
#		define	NOPROGRESS	010	/* zero-progress state set */
	struct arcp ins;	/* chain of inarcs pointing here */
	TBufStringMarker	lastseen ;	/* last entered on arrival here */
	struct sset **outs;	/* outarc vector indexed by color */
	struct arcp *inchain;	/* chain-pointer vector for outarcs */
};

struct dfa {
	int nssets;		/* size of cache */
	int nssused;		/* how many entries occupied yet */
	int nstates;		/* number of states */
	int ncolors;		/* length of outarc and inchain vectors */
	int wordsper;		/* length of state-set bitvectors */
	struct sset *ssets;	/* state-set cache */
	unsigned *statesarea;	/* bitvector storage */
	unsigned *work;		/* pointer to work area within statesarea */
	struct sset **outsarea;	/* outarc-vector storage */
	struct arcp *incarea;	/* inchain storage */
	struct cnfa *cnfa;
	struct colormap *cm;
	TBufStringMarker	lastpost ;		/* location of last cache-flushed success */
	TBufStringMarker	lastnopr ;		/* location of last cache-flushed NOPROGRESS */
	struct sset *search;	/* replacement-search-pointer memory */
	int cptsmalloced;	/* were the areas individually malloced? */
	char *mallocarea;	/* self, or master malloced area, or NULL */
};

#define	WORK	1		/* number of work bitvectors needed */

/* setup for non-malloc allocation for small cases */
#define	FEWSTATES	20	/* must be less than UBITS */
#define	FEWCOLORS	15
struct smalldfa {
	struct dfa dfa;
	struct sset ssets[FEWSTATES*2];
	unsigned statesarea[FEWSTATES*2 + WORK];
	struct sset *outsarea[FEWSTATES*2 * FEWCOLORS];
	struct arcp incarea[FEWSTATES*2 * FEWCOLORS];
};
#define	DOMALLOC	((struct smalldfa *)NULL)	/* force malloc */



/* internal variables, bundled for easy passing around */
struct vars {
	regex_t*			re ;
	struct guts*		g ;
	int					eflags ;		/* copies of arguments */
	size_t				nmatch ;
	regmatch_t*			pmatch ;
	rm_detail_t*		details ;
	TBufStringMarker	start ;		/* start of string */
	TBufStringMarker	stop ;		/* just past end of string */
	int					err ;		/* error code if any (0 none) */
	regoff_t*			mem;		/* memory vector for backtracking */
	struct smalldfa		dfa1 ;
	struct smalldfa		dfa2 ;
} ;
#define	VISERR(vv)	((vv)->err != 0)	/* have we seen an error yet? */
#define	ISERR()	VISERR(v)
#define	VERR(vv,e)	(((vv)->err) ? (vv)->err : ((vv)->err = (e)))
#define	ERR(e)	VERR(v, e)		/* record an error */
#define	NOERR()	{if (ISERR()) return v->err;}	/* if error seen, return it */
#define	OFF(p)	(TBufStringMarker_GetOffset (&p))
#define	LOFF(p)	((long)OFF(p))

/*
 * forward declarations
 */
/* =====^!^===== begin forwards =====^!^===== */
/* automatically gathered by fwd; do not hand-edit */
/* === regexec.c === */
int exec _ANSI_ARGS_((regex_t *, CONST chr *, size_t, rm_detail_t *, size_t, regmatch_t [], int));
static int find_2 _ANSI_ARGS_((struct vars *, struct cnfa *, struct colormap *));
static int cfind_2 _ANSI_ARGS_((struct vars *, struct cnfa *, struct colormap *));
static int cfindloop_2 _ANSI_ARGS_((struct vars *, struct cnfa *, struct colormap *, struct dfa *, struct dfa *, TBufStringMarker*));
static VOID zapsubs _ANSI_ARGS_((regmatch_t *, size_t));
static VOID zapmem _ANSI_ARGS_((struct vars *, struct subre *));
static VOID subset_2 _ANSI_ARGS_((struct vars *, struct subre *, TBufStringMarker, TBufStringMarker));
static int dissect_2 _ANSI_ARGS_((struct vars *, struct subre *, TBufStringMarker, TBufStringMarker)) ;
static int condissect_2 _ANSI_ARGS_((struct vars *, struct subre *, TBufStringMarker, TBufStringMarker)) ;
static int altdissect_2 _ANSI_ARGS_((struct vars *, struct subre *, TBufStringMarker, TBufStringMarker)) ;
static int cdissect_2 _ANSI_ARGS_((struct vars *, struct subre *, TBufStringMarker, TBufStringMarker)) ;
static int ccondissect_2 _ANSI_ARGS_((struct vars *, struct subre *, TBufStringMarker, TBufStringMarker));
static int crevdissect_2 _ANSI_ARGS_((struct vars *, struct subre *, TBufStringMarker, TBufStringMarker));
static int cbrdissect_2 _ANSI_ARGS_((struct vars *, struct subre *, TBufStringMarker, TBufStringMarker));
static int caltdissect_2 _ANSI_ARGS_((struct vars *, struct subre *, TBufStringMarker, TBufStringMarker));
/* === rege_dfa_2.c === */
static int	longest_2 _ANSI_ARGS_((struct vars *, struct dfa *, TBufStringMarker, TBufStringMarker, int*, TBufStringMarker* pRetval)) ;

static int	shortest_2 (struct vars*, struct dfa*, TBufStringMarker, TBufStringMarker, TBufStringMarker, TBufStringMarker*, int*, TBufStringMarker*	pRetval) ;
static int	lastcold_2 _ANSI_ARGS_((struct vars *, struct dfa *, TBufStringMarker*));
static struct dfa *newdfa _ANSI_ARGS_((struct vars *, struct cnfa *, struct colormap *, struct smalldfa *));
static VOID freedfa _ANSI_ARGS_((struct dfa *));
static unsigned hash _ANSI_ARGS_((unsigned *, int));
static struct sset *initialize_2 _ANSI_ARGS_((struct vars *, struct dfa *, TBufStringMarker));
static struct sset *miss_2 _ANSI_ARGS_((struct vars *, struct dfa *, struct sset *, pcolor, TBufStringMarker, TBufStringMarker));
static int lacon_2 _ANSI_ARGS_((struct vars *, struct cnfa *, TBufStringMarker, pcolor));
static struct sset *getvacant_2 _ANSI_ARGS_((struct vars *, struct dfa *, TBufStringMarker, TBufStringMarker)) ;
static struct sset *pickss_2 _ANSI_ARGS_((struct vars *, struct dfa *, TBufStringMarker, TBufStringMarker)) ;
/* automatically gathered by fwd; do not hand-edit */
/* =====^!^===== end forwards =====^!^===== */

int cmp _ANSI_ARGS_((CONST chr *, CONST chr *, size_t));
int casecmp _ANSI_ARGS_((CONST chr *, CONST chr *, size_t));
static int cmp_2 _ANSI_ARGS_((CONST TBufStringMarker*, CONST TBufStringMarker*, size_t));
static int casecmp_2 _ANSI_ARGS_((CONST TBufStringMarker*, CONST TBufStringMarker*, size_t));

/*
 - exec - match regular expression
 ^ int exec(regex_t *, CONST chr *, size_t, rm_detail_t *,
 ^					size_t, regmatch_t [], int);
 */
int
exec_2 (
	regex_t*				re,
	CONST TBufStringMarker	strstart,
	CONST TBufStringMarker	strstop,
	rm_detail_t*			details,
	size_t					nmatch,
	regmatch_t				pmatch [],
	int						flags)
{
	struct vars var;
	register struct vars *v = &var;
	int st;
	size_t n;
	int backref;
#	define	LOCALMAT	20
	regmatch_t mat[LOCALMAT];
#	define	LOCALMEM	40
	regoff_t mem[LOCALMEM];

	/* sanity checks */
	if (re == NULL || /*string == NULL ||*/ re->re_magic != REMAGIC)
		return REG_INVARG;
	if (re->re_csize != sizeof(chr))
		return REG_MIXED;

	/* setup */
	v->re = re;
	v->g = (struct guts *)re->re_guts;
	if ((v->g->cflags&REG_EXPECT) && details == NULL)
		return REG_INVARG;
	if (v->g->info&REG_UIMPOSSIBLE)
		return REG_NOMATCH;
	backref = (v->g->info&REG_UBACKREF) ? 1 : 0;
	v->eflags = flags;
	if (v->g->cflags&REG_NOSUB)
		nmatch = 0;		/* override client */
	v->nmatch = nmatch;
	if (backref) {
		/* need work area */
		if (v->g->nsub + 1 <= LOCALMAT)
			v->pmatch = mat;
		else
			v->pmatch = (regmatch_t *)MALLOC((v->g->nsub + 1) *
							sizeof(regmatch_t));
		if (v->pmatch == NULL)
			return REG_ESPACE;
		v->nmatch = v->g->nsub + 1;
	} else
		v->pmatch = pmatch;
	v->details	= details;
	v->start	= strstart ;
	v->stop		= strstop ;
	v->err		= 0;
	if (backref) {
		/* need retry memory */
		assert(v->g->ntree >= 0);
		n = (size_t)v->g->ntree;
		if (n <= LOCALMEM)
			v->mem = mem;
		else
			v->mem = (regoff_t *)MALLOC(n*sizeof(regoff_t));
		if (v->mem == NULL) {
			if (v->pmatch != pmatch && v->pmatch != mat)
				FREE(v->pmatch);
			return REG_ESPACE;
		}
	} else
		v->mem = NULL;

	/* do it */
	assert(v->g->tree != NULL);
	if (backref)
		st = cfind_2 (v, &v->g->tree->cnfa, &v->g->cmap);
	else
		st = find_2 (v, &v->g->tree->cnfa, &v->g->cmap);

	/* copy (portion of) match vector over if necessary */
	if (st == REG_OKAY && v->pmatch != pmatch && nmatch > 0) {
		zapsubs(pmatch, nmatch);
		n = (nmatch < v->nmatch) ? nmatch : v->nmatch;
		memcpy(VS(pmatch), VS(v->pmatch), n*sizeof(regmatch_t));
	}

	/* clean up */
	if (v->pmatch != pmatch && v->pmatch != mat)
		FREE(v->pmatch);
	if (v->mem != NULL && v->mem != mem)
		FREE(v->mem);
	return st;
}

/*
 - find - find a match for the main NFA (no-complications case)
 ^ static int find(struct vars *, struct cnfa *, struct colormap *);
 */
static int
find_2 (
	register struct vars*		v,
	register struct cnfa*		cnfa,
	register struct colormap*	cm)
{
	struct dfa*			s ;
	struct dfa*			d ;
	TBufStringMarker	begin ;
	TBufStringMarker	end ;
	TBufStringMarker	cold ;
	TBufStringMarker	open;		/* open and close of range of possible starts */
	TBufStringMarker	close ;
	int hitend;
	int shorter = (v->g->tree->flags&SHORTER) ? 1 : 0;
	int	nres ;

	TBufStringMarker_SetVoid (&end) ;

	/* first, a shot with the search RE */
	s = newdfa(v, &v->g->search, cm, &v->dfa1);
	assert(!(ISERR() && s != NULL));
	NOERR();
	MDEBUG(("\nsearch at %ld\n", LOFF(v->start)));
	TBufStringMarker_SetVoid (&cold) ;
	shortest_2(v, s, v->start, v->start, v->stop, &cold, (int *)NULL, &close) ;
	freedfa(s);
	NOERR();
	if (v->g->cflags&REG_EXPECT) {
		assert(v->details != NULL);
		if (!TBufStringMarker_Voidp (&cold))
			v->details->rm_extend.rm_so = OFF(cold) ;
		else
			v->details->rm_extend.rm_so = OFF(v->stop);
		v->details->rm_extend.rm_eo = OFF(v->stop);	/* unknown */
	}
	if (TBufStringMarker_Voidp (&close))		/* not found */
		return REG_NOMATCH;
	if (v->nmatch == 0)		/* found, don't need exact location */
		return REG_OKAY;

	/* find starting point and match */
	assert(!TBufStringMarker_Voidp (&cold));
	open = cold;
	TBufStringMarker_SetVoid (&cold) ;
	MDEBUG(("between %ld and %ld\n", LOFF(open), LOFF(close)));
	d = newdfa(v, cnfa, cm, &v->dfa1);
	assert(!(ISERR() && d != NULL));
	NOERR();
	for (begin = open; OFF(begin) <= OFF(close) ; TBufStringMarker_Forward (&begin, 1)) {
		MDEBUG(("\nfind trying at %ld\n", LOFF(begin)));
		if (shorter)
			nres	= shortest_2 (v, d, begin, begin, v->stop, NULL, &hitend, &end) ;
		else
			nres	= longest_2 (v, d, begin, v->stop, &hitend, &end);
		NOERR();
		if (hitend && TBufStringMarker_Voidp (&cold))
			cold = begin;
		if (!TBufStringMarker_Voidp (&end))
			break;		/* NOTE BREAK OUT */
	}
	assert(!TBufStringMarker_Voidp (&end));		/* search RE succeeded so loop should */
	freedfa(d);

	/* and pin down details */
	assert(v->nmatch > 0);
	v->pmatch[0].rm_so = OFF(begin);
	v->pmatch[0].rm_eo = OFF(end);
	if (v->g->cflags&REG_EXPECT) {
		if (!TBufStringMarker_Voidp (&cold))
			v->details->rm_extend.rm_so = OFF(cold);
		else
			v->details->rm_extend.rm_so = OFF(v->stop);
		v->details->rm_extend.rm_eo = OFF(v->stop);	/* unknown */
	}
	if (v->nmatch == 1)		/* no need for submatches */
		return REG_OKAY;

	/* submatches */
	zapsubs(v->pmatch, v->nmatch);
	return dissect_2(v, v->g->tree, begin, end);
}

/*
 - cfind - find a match for the main NFA (with complications)
 ^ static int cfind(struct vars *, struct cnfa *, struct colormap *);
 */
static int
cfind_2 (
	register struct vars*		v,
	register struct cnfa*		cnfa,
	register struct colormap*	cm)
{
	struct dfa*			s ;
	struct dfa*			d ;
	TBufStringMarker	cold ;
	int					ret ;

	s = newdfa(v, &v->g->search, cm, &v->dfa1);
	NOERR();
	d = newdfa(v, cnfa, cm, &v->dfa2);
	if (ISERR()) {
		assert(d == NULL);
		freedfa(s);
		return v->err;
	}

	ret = cfindloop_2 (v, cnfa, cm, d, s, &cold);

	freedfa(d);
	freedfa(s);
	NOERR();
	if (v->g->cflags&REG_EXPECT) {
		assert(v->details != NULL);
		if (!TBufStringMarker_Voidp (&cold))
			v->details->rm_extend.rm_so = OFF(cold);
		else
			v->details->rm_extend.rm_so = OFF(v->stop);
		v->details->rm_extend.rm_eo = OFF(v->stop);	/* unknown */
	}
	return ret;
}

/*
 - cfindloop - the heart of cfind
 ^ static int cfindloop(struct vars *, struct cnfa *, struct colormap *,
 ^	struct dfa *, struct dfa *, chr **);
 */
static int
cfindloop_2 (
	register struct vars*		v,
	register struct cnfa*		cnfa,
	register struct colormap*	cm,
	register struct dfa*		d,
	register struct dfa*		s,
	register TBufStringMarker*	coldp)			/* where to put coldstart pointer */
{
	TBufStringMarker	begin ;
	TBufStringMarker	end ;
	TBufStringMarker	cold ;
	TBufStringMarker	open ;	/* open and close of range of possible starts */
	TBufStringMarker	close ;
	TBufStringMarker	estart ;
	TBufStringMarker	estop ;
	int					er ;
	int					shorter	= v->g->tree->flags & SHORTER ;
	int					hitend ;

	assert (d != NULL && s != NULL) ;
	TBufStringMarker_SetVoid (&cold) ;
	close	= v->start ;
	do {
		MDEBUG(("\ncsearch at %ld\n", LOFF(close)));
		shortest_2 (v, s, close, close, v->stop, &cold, (int *)NULL, &close) ;
		if (TBufStringMarker_Voidp (&close))
			break;				/* NOTE BREAK */
		assert (!TBufStringMarker_Voidp (&cold)) ;
		open	= cold ;
		TBufStringMarker_SetVoid (&cold) ;
		MDEBUG(("cbetween %ld and %ld\n", LOFF(open), LOFF(close))) ;
		for (begin = open; OFF(begin) <= OFF(close) ; TBufStringMarker_Forward (&begin, 1)) {
			MDEBUG(("\ncfind trying at %ld\n", LOFF(begin)));
			estart	= begin ;
			estop	= v->stop ;
			for ( ; ; ) {
				if (shorter) {
					shortest_2 (v, d, begin, estart, estop, (TBufStringMarker*)NULL, &hitend, &end) ;
				} else {
					longest_2 (v, d, begin, estop, &hitend, &end) ;
				}
				if (hitend && TBufStringMarker_Voidp (&cold))
					cold	= begin ;
				if (TBufStringMarker_Voidp (&end))
					break ;		/* NOTE BREAK OUT */
				MDEBUG(("tentative end %ld\n", LOFF(end)));
				zapsubs(v->pmatch, v->nmatch) ;
				zapmem(v, v->g->tree) ;
				er = cdissect_2 (v, v->g->tree, begin, end) ;
				if (er == REG_OKAY) {
					if (v->nmatch > 0) {
						v->pmatch[0].rm_so = OFF(begin);
						v->pmatch[0].rm_eo = OFF(end);
					}
					*coldp = cold;
					return REG_OKAY;
				}
				if (er != REG_NOMATCH) {
					ERR(er);
					return er;
				}
				if ((shorter) ? OFF(end) == OFF(estop) : OFF(end) == OFF(begin)) {
					/* no point in trying again */
					*coldp = cold;
					return REG_NOMATCH;
				}
				/* go around and try again */
				if (shorter) {
					estart	= end ;
					TBufStringMarker_Forward (&estart, 1) ;
				} else {
					estop	= end ;
					TBufStringMarker_Backward (&estop, 1) ;
				}
			}
		}
	} while (OFF(close) < OFF(v->stop)) ;

	*coldp = cold;
	return REG_NOMATCH;
}

/*
 - zapsubs - initialize the subexpression matches to "no match"
 ^ static VOID zapsubs(regmatch_t *, size_t);
 */
static VOID
zapsubs(p, n)
regmatch_t *p;
size_t n;
{
	size_t i;

	for (i = n-1; i > 0; i--) {
		p[i].rm_so = -1;
		p[i].rm_eo = -1;
	}
}

/*
 - zapmem - initialize the retry memory of a subtree to zeros
 ^ static VOID zapmem(struct vars *, struct subre *);
 */
static VOID
zapmem(v, t)
struct vars *v;
struct subre *t;
{
	if (t == NULL)
		return;

	assert(v->mem != NULL);
	v->mem[t->retry] = 0;
	if (t->op == '(') {
		assert(t->subno > 0);
		v->pmatch[t->subno].rm_so = -1;
		v->pmatch[t->subno].rm_eo = -1;
	}

	if (t->left != NULL)
		zapmem(v, t->left);
	if (t->right != NULL)
		zapmem(v, t->right);
}

/*
 - subset - set any subexpression relevant to a successful subre
 ^ static VOID subset(struct vars *, struct subre *, chr *, chr *);
 */
static VOID
subset_2 (
	register struct vars*	v,
	register struct subre*	sub,
	TBufStringMarker		begin,
	TBufStringMarker		end)
{
	register int n = sub->subno ;

	assert(n > 0);
	if ((size_t)n >= v->nmatch)
		return;

	MDEBUG(("setting %d\n", n));
	v->pmatch[n].rm_so	= OFF(begin) ;
	v->pmatch[n].rm_eo	= OFF(end) ;
}

/*
 - dissect - determine subexpression matches (uncomplicated case)
 ^ static int dissect(struct vars *, struct subre *, chr *, chr *);
 */
static int			/* regexec return code */
dissect_2 (
	register struct vars*		v,
	register struct subre*		t,
	TBufStringMarker	begin,	/* beginning of relevant substring */
	TBufStringMarker	end)	/* end of same */
{
	assert(t != NULL);
	MDEBUG(("dissect %ld-%ld\n", LOFF(begin), LOFF(end)));

	switch (t->op) {
	case '=':		/* terminal node */
		assert(t->left == NULL && t->right == NULL);
		return	REG_OKAY;	/* no action, parent did the work */
		break;
	case '|':		/* alternation */
		assert(t->left != NULL);
		return	altdissect_2 (v, t, begin, end) ;
		break;
	case 'b':		/* back ref -- shouldn't be calling us! */
		return REG_ASSERT;
		break;
	case '.':		/* concatenation */
		assert (t->left != NULL && t->right != NULL);
		return	condissect_2 (v, t, begin, end) ;
		break;
	case '(':		/* capturing */
		assert(t->left != NULL && t->right == NULL);
		assert(t->subno > 0);
		subset_2 (v, t, begin, end) ;
		return	dissect_2 (v, t->left, begin, end) ;
		break;
	default:
		return REG_ASSERT;
		break;
	}
}

/*
 - condissect - determine concatenation subexpression matches (uncomplicated)
 ^ static int condissect(struct vars *, struct subre *, chr *, chr *);
 */
static int			/* regexec return code */
condissect_2 (
	register struct vars*	v,
	register struct subre*	t,
	TBufStringMarker		begin,			/* beginning of relevant substring */
	TBufStringMarker		end)			/* end of same */
{
	register struct dfa*	d ;
	register struct dfa*	d2 ;
	TBufStringMarker		mid ;
	register int			i;
	register int			shorter = (t->left->flags&SHORTER) ? 1 : 0;
	TBufStringMarker		stop	= (shorter) ? end : begin ;
	TBufStringMarker		tmp ;

	assert(t->op == '.');
	assert(t->left != NULL && t->left->cnfa.nstates > 0);
	assert(t->right != NULL && t->right->cnfa.nstates > 0);

	d = newdfa(v, &t->left->cnfa, &v->g->cmap, &v->dfa1);
	NOERR();
	d2 = newdfa(v, &t->right->cnfa, &v->g->cmap, &v->dfa2);
	if (ISERR()) {
		assert(d2 == NULL);
		freedfa(d);
		return v->err;
	}

	/* pick a tentative midpoint */
	if (shorter) {
		shortest_2 (v, d, begin, begin, end, (TBufStringMarker*)NULL, (int *)NULL, &mid) ;
	} else {
		longest_2 (v, d, begin, end, (int *)NULL, &mid) ;
	}
	if (TBufStringMarker_Voidp (&mid)) {
		freedfa (d) ;
		freedfa (d2) ;
		return	REG_ASSERT;
	}
	MDEBUG(("tentative midpoint %ld\n", LOFF(mid)));

	/* iterate until satisfaction or failure */
	while (longest_2 (v, d2, mid, end, (int *)NULL, &tmp), OFF(tmp) != OFF(end)) {
		/* that midpoint didn't work, find a new one */
		if (OFF(mid) == OFF(stop)) {
			/* all possibilities exhausted! */
			MDEBUG(("no midpoint!\n"));
			freedfa (d) ;
			freedfa (d2) ;
			return	REG_ASSERT ;
		}
		if (shorter) {
			TBufStringMarker_Forward (&mid, 1) ;
			shortest_2 (v, d, begin, mid, end, (TBufStringMarker*)NULL, (int *)NULL, &mid) ;
		} else {
			TBufStringMarker_Backward (&mid, 1) ;
			longest_2 (v, d, begin, mid, (int *)NULL, &mid) ;
		}
		if (TBufStringMarker_Voidp (&mid)) {
			/* failed to find a new one! */
			MDEBUG (("failed midpoint!\n"));
			freedfa (d) ;
			freedfa (d2) ;
			return	REG_ASSERT ;
		}
		MDEBUG(("new midpoint %ld\n", LOFF(mid)));
	}

	/* satisfaction */
	MDEBUG(("successful\n"));
	freedfa (d) ;
	freedfa (d2) ;
	i	= dissect_2 (v, t->left, begin, mid) ;
	if (i != REG_OKAY)
		return	i ;
	return	dissect_2 (v, t->right, mid, end) ;
}

/*
 - altdissect - determine alternative subexpression matches (uncomplicated)
 ^ static int altdissect(struct vars *, struct subre *, chr *, chr *);
 */
static int			/* regexec return code */
altdissect_2 (
	register struct vars*	v,
	register struct subre*	t,
	TBufStringMarker		begin,			/* beginning of relevant substring */
	TBufStringMarker		end)			/* end of same */
{
	struct dfa*			d ;
	int					i ;
	TBufStringMarker	ret ;

	assert(t != NULL) ;
	assert(t->op == '|') ;

	for (i = 0; t != NULL; t = t->right, i++) {
		MDEBUG(("trying %dth\n", i));
		assert (t->left != NULL && t->left->cnfa.nstates > 0) ;
		d	= newdfa(v, &t->left->cnfa, &v->g->cmap, &v->dfa1);
		if (ISERR())
			return v->err;
		TBufStringMarker_SetVoid (&ret) ;
		longest_2 (v, d, begin, end, (int *)NULL, &ret) ;
		if (OFF(ret) == OFF(end)) {
			MDEBUG(("success\n"));
			freedfa(d);
			return	dissect_2 (v, t->left, begin, end) ;
		}
		freedfa (d) ;
	}
	return	REG_ASSERT;	/* none of them matched?!? */
}

/*
 - cdissect - determine subexpression matches (with complications)
 * The retry memory stores the offset of the trial midpoint from begin, 
 * plus 1 so that 0 uniquely means "clean slate".
 ^ static int cdissect(struct vars *, struct subre *, chr *, chr *);
 */
static int			/* regexec return code */
cdissect_2 (
	register struct vars*	v,
	register struct subre*	t,
	TBufStringMarker		begin,			/* beginning of relevant substring */
	TBufStringMarker		end)			/* end of same */
{
	register int	er ;

	assert(t != NULL);
	MDEBUG(("cdissect %ld-%ld %c\n", LOFF(begin), LOFF(end), t->op));

	switch (t->op) {
	case '=':		/* terminal node */
		assert(t->left == NULL && t->right == NULL);
		return	REG_OKAY;	/* no action, parent did the work */
		break;
	case '|':		/* alternation */
		assert(t->left != NULL);
		return	caltdissect_2 (v, t, begin, end) ;
		break;
	case 'b':		/* back ref -- shouldn't be calling us! */
		assert(t->left == NULL && t->right == NULL);
		return	cbrdissect_2 (v, t, begin, end) ;
		break;
	case '.':		/* concatenation */
		assert(t->left != NULL && t->right != NULL);
		return	ccondissect_2 (v, t, begin, end) ;
		break;
	case '(':		/* capturing */
		assert(t->left != NULL && t->right == NULL);
		assert(t->subno > 0);
		er	= cdissect_2 (v, t->left, begin, end);
		if (er == REG_OKAY)
			subset_2 (v, t, begin, end) ;
		return er;
		break;
	default:
		return REG_ASSERT;
		break;
	}
}

/*
 - ccondissect - concatenation subexpression matches (with complications)
 * The retry memory stores the offset of the trial midpoint from begin, 
 * plus 1 so that 0 uniquely means "clean slate".
 ^ static int ccondissect(struct vars *, struct subre *, chr *, chr *);
 */
static int			/* regexec return code */
ccondissect_2 (
	register struct vars*	v,
	register struct subre*	t,
	TBufStringMarker		begin,			/* beginning of relevant substring */
	TBufStringMarker		end)			/* end of same */
{
	register struct dfa*	d ;
	register struct dfa*	d2 ;
	TBufStringMarker		mid ;
	register int			er ;

	assert(t->op == '.');
	assert(t->left != NULL && t->left->cnfa.nstates > 0);
	assert(t->right != NULL && t->right->cnfa.nstates > 0);

	if (t->left->flags & SHORTER)		/* reverse scan */
		return	crevdissect_2 (v, t, begin, end) ;

	d = newdfa(v, &t->left->cnfa, &v->g->cmap, DOMALLOC);
	if (ISERR())
		return v->err;
	d2 = newdfa(v, &t->right->cnfa, &v->g->cmap, DOMALLOC);
	if (ISERR()) {
		freedfa(d);
		return v->err;
	}
	MDEBUG(("cconcat %d\n", t->retry));

	/* pick a tentative midpoint */
	if (v->mem[t->retry] == 0) {
		longest_2 (v, d, begin, end, (int *)NULL, &mid) ;
		if (TBufStringMarker_Voidp (&mid)) {
			freedfa (d) ;
			freedfa (d2) ;
			return	REG_NOMATCH ;
		}
		MDEBUG(("tentative midpoint %ld\n", LOFF(mid)));
		v->mem[t->retry] = (OFF(mid) - OFF(begin)) + 1 ;
	} else {
		mid	= begin ;
		TBufStringMarker_Forward (&mid, v->mem[t->retry] - 1) ;
		MDEBUG(("working midpoint %ld\n", LOFF(mid)));
	}

	/* iterate until satisfaction or failure */
	for (;;) {
		/* try this midpoint on for size */
		er = cdissect_2 (v, t->left, begin, mid) ;
		if (er == REG_OKAY) {
			TBufStringMarker	tmp ;
			longest_2 (v, d2, mid, end, (int *)NULL, &tmp) ;
			if (OFF(tmp) == OFF(end) &&
				(er = cdissect_2 (v, t->right, mid, end)) == REG_OKAY) {
				break;			/* NOTE BREAK OUT */
			}
		}
		if (er != REG_OKAY && er != REG_NOMATCH) {
			freedfa (d) ;
			freedfa (d2) ;
			return	er ;
		}

		/* that midpoint didn't work, find a new one */
		if (OFF(mid) == OFF(begin)) {
			/* all possibilities exhausted */
			MDEBUG(("%d no midpoint\n", t->retry));
			freedfa (d) ;
			freedfa (d2) ;
			return	REG_NOMATCH ;
		}
		TBufStringMarker_Backward (&mid, 1) ;
		longest_2 (v, d, begin, mid, (int *)NULL, &mid) ;
		if (TBufStringMarker_Voidp (&mid)) {
			/* failed to find a new one */
			MDEBUG(("%d failed midpoint\n", t->retry));
			freedfa (d) ;
			freedfa (d2) ;
			return	REG_NOMATCH ;
		}
		MDEBUG(("%d: new midpoint %ld\n", t->retry, LOFF(mid)));
		v->mem[t->retry] = (OFF(mid) - OFF(begin)) + 1 ;
		zapmem(v, t->left);
		zapmem(v, t->right);
	}

	/* satisfaction */
	MDEBUG(("successful\n"));
	freedfa(d);
	freedfa(d2);
	return REG_OKAY;
}

/*
 - crevdissect - determine backref shortest-first subexpression matches
 * The retry memory stores the offset of the trial midpoint from begin, 
 * plus 1 so that 0 uniquely means "clean slate".
 ^ static int crevdissect(struct vars *, struct subre *, chr *, chr *);
 */
static int			/* regexec return code */
crevdissect_2 (
	register struct vars*	v,
	register struct subre*	t,
	TBufStringMarker		begin,			/* beginning of relevant substring */
	TBufStringMarker		end)			/* end of same */
{
	register struct dfa*	d ;
	register struct dfa*	d2 ;
	TBufStringMarker		mid ;
	register int			er ;

	assert(t->op == '.');
	assert(t->left != NULL && t->left->cnfa.nstates > 0);
	assert(t->right != NULL && t->right->cnfa.nstates > 0);
	assert(t->left->flags&SHORTER);

	/* concatenation -- need to split the substring between parts */
	d = newdfa(v, &t->left->cnfa, &v->g->cmap, DOMALLOC);
	if (ISERR())
		return v->err;
	d2 = newdfa(v, &t->right->cnfa, &v->g->cmap, DOMALLOC);
	if (ISERR()) {
		freedfa(d);
		return v->err;
	}
	MDEBUG(("crev %d\n", t->retry));

	/* pick a tentative midpoint */
	if (v->mem[t->retry] == 0) {
		shortest_2 (v, d, begin, begin, end, (TBufStringMarker*)NULL, (int *)NULL, &mid) ;
		if (TBufStringMarker_Voidp (&mid)) {
			freedfa (d) ;
			freedfa (d2) ;
			return	REG_NOMATCH ;
		}
		MDEBUG(("tentative midpoint %ld\n", LOFF(mid)));
		v->mem[t->retry]	= (OFF(mid) - OFF(begin)) + 1 ;
	} else {
		mid	= begin ;
		TBufStringMarker_Forward (&mid, v->mem[t->retry] - 1) ;
		MDEBUG(("working midpoint %ld\n", LOFF(mid)));
	}

	/* iterate until satisfaction or failure */
	for ( ; ; ) {
		/* try this midpoint on for size */
		er	= cdissect_2 (v, t->left, begin, mid) ;
		if (er == REG_OKAY) {
			TBufStringMarker	tmp ;
			longest_2 (v, d2, mid, end, (int *)NULL, &tmp) ;
			if (OFF(tmp) == OFF(end) &&
				(er = cdissect_2 (v, t->right, mid, end)) == REG_OKAY) {
				break;			/* NOTE BREAK OUT */
			}
		}
		if (er != REG_OKAY && er != REG_NOMATCH) {
			freedfa (d) ;
			freedfa (d2) ;
			return	er ;
		}

		/* that midpoint didn't work, find a new one */
		if (OFF(mid) == OFF(end)) {
			/* all possibilities exhausted */
			MDEBUG(("%d no midpoint\n", t->retry));
			freedfa (d) ;
			freedfa (d2) ;
			return	REG_NOMATCH ;
		}
		TBufStringMarker_Forward (&mid, 1) ;
		shortest_2 (v, d, begin, mid, end, (TBufStringMarker*)NULL, (int *)NULL, &mid) ;
		if (TBufStringMarker_Voidp (&mid)) {
			/* failed to find a new one */
			MDEBUG(("%d failed midpoint\n", t->retry));
			freedfa (d) ;
			freedfa (d2) ;
			return	REG_NOMATCH ;
		}
		MDEBUG(("%d: new midpoint %ld\n", t->retry, LOFF(mid)));
		v->mem[t->retry] = (OFF(mid) - OFF(begin)) + 1 ;
		zapmem (v, t->left) ;
		zapmem (v, t->right) ;
	}

	/* satisfaction */
	MDEBUG(("successful\n"));
	freedfa (d) ;
	freedfa (d2) ;
	return	REG_OKAY ;
}

/*
 - cbrdissect - determine backref subexpression matches
 ^ static int cbrdissect(struct vars *, struct subre *, chr *, chr *);
 */
static int			/* regexec return code */
cbrdissect_2 (
	register struct vars*	v,
	register struct subre*	t,
	TBufStringMarker		begin,			/* beginning of relevant substring */
	TBufStringMarker		end)			/* end of same */
{
	register int		i ;
	register int		n	= t->subno ;
	register size_t		len ;
	TBufStringMarker	paren ;
	TBufStringMarker	p ;
	TBufStringMarker	stop ;
	register int		min	= t->min ;
	register int		max	= t->max ;
	register int		(*comparep)(CONST TBufStringMarker*, CONST TBufStringMarker*, size_t) ;

	assert(t != NULL);
	assert(t->op == 'b') ;
	assert(n >= 0);
	assert((size_t)n < v->nmatch);

	MDEBUG(("cbackref n%d %d{%d-%d}\n", t->retry, n, min, max));

	if (v->pmatch[n].rm_so == -1)
		return	REG_NOMATCH ;

	paren	= v->start ;
	TBufStringMarker_Forward (&paren, v->pmatch[n].rm_so) ;
	len		= v->pmatch[n].rm_eo - v->pmatch[n].rm_so;

	/* no room to maneuver -- retries are pointless */
	if (v->mem[t->retry])
		return	REG_NOMATCH ;
	v->mem [t->retry]	= 1 ;

	/* special-case zero-length string */
	if (len == 0) {
		if (OFF(begin) == OFF(end))
			return	REG_OKAY ;
		return	REG_NOMATCH ;
	}

	/* and too-short string */
	assert (OFF(end) >= OFF(begin)) ;
	if ((size_t)(OFF(end) - OFF(begin)) < len)
		return	REG_NOMATCH ;

	stop	= end ;
	TBufStringMarker_Backward (&stop, len) ;

	/* count occurrences */
	i		= 0 ;
	if (*v->g->compare == casecmp) {
		comparep	= casecmp_2 ;
	} else {
		comparep	= cmp_2 ;
	}
	for (p = begin ; OFF(p) <= OFF(stop) && (i < max || max == INFINITY); TBufStringMarker_Forward (&p, len)) {
		if ((comparep)(&paren, &p, len) != 0)
			break ;
		i++ ;
	}
	MDEBUG(("cbackref found %d\n", i)) ;

	/* and sort it out */
	if (OFF(p) != OFF(end))			/* didn't consume all of it */
		return	REG_NOMATCH ;
	if (min <= i && (i <= max || max == INFINITY))
		return	REG_OKAY;
	return	REG_NOMATCH;		/* out of range */
}

/*
 - caltdissect - determine alternative subexpression matches (w. complications)
 ^ static int caltdissect(struct vars *, struct subre *, chr *, chr *);
 */
static int			/* regexec return code */
caltdissect_2 (
	register struct vars*	v,
	register struct subre*	t,
	TBufStringMarker		begin,			/* beginning of relevant substring */
	TBufStringMarker		end)			/* end of same */
{
	register struct dfa*	d ;
	register int			er ;
#	define	UNTRIED	0	/* not yet tried at all */
#	define	TRYING	1	/* top matched, trying submatches */
#	define	TRIED	2	/* top didn't match or submatches exhausted */

	if (t == NULL)
		return	REG_NOMATCH ;
	assert(t->op == '|');
	if (v->mem[t->retry] == TRIED)
		return	caltdissect_2 (v, t->right, begin, end) ;

	MDEBUG(("calt n%d\n", t->retry));
	assert(t->left != NULL);

	if (v->mem[t->retry] == UNTRIED) {
		TBufStringMarker	tmp ;
		d = newdfa(v, &t->left->cnfa, &v->g->cmap, DOMALLOC);
		if (ISERR())
			return v->err;
		longest_2 (v, d, begin, end, (int *)NULL, &tmp) ;
		if (OFF(tmp) != OFF(end)) {
			freedfa (d) ;
			v->mem[t->retry] = TRIED ;
			return	caltdissect_2 (v, t->right, begin, end) ;
		}
		freedfa(d);
		MDEBUG(("calt matched\n"));
		v->mem[t->retry] = TRYING;
	}

	er = cdissect_2 (v, t->left, begin, end);
	if (er != REG_NOMATCH)
		return er;

	v->mem[t->retry] = TRIED;
	return	caltdissect_2 (v, t->right, begin, end) ;
}

/*
 - cmp - chr-substring compare
 * Backrefs need this.  It should preferably be efficient.
 * Note that it does not need to report anything except equal/unequal.
 * Note also that the length is exact, and the comparison should not
 * stop at embedded NULs!
 ^ static int cmp(CONST chr *, CONST chr *, size_t);
 */
static int			/* 0 for equal, nonzero for unequal */
cmp_2 (
	register CONST TBufStringMarker*	xp,
	register CONST TBufStringMarker*	yp,
	register size_t						len)	/* exact length of comparison */
{
	TBufStringMarker	x	= *xp ;
	TBufStringMarker	y	= *yp ;
	while (len > 0) {
		if (TBufStringMarker_GetChar (&x) != TBufStringMarker_GetChar (&y))
			return	1 ;
		TBufStringMarker_Forward (&x, 1) ;
		TBufStringMarker_Forward (&y, 1) ;
		len	-- ;
	}
	return	0 ;
}

/*
 - casecmp - case-independent chr-substring compare
 * REG_ICASE backrefs need this.  It should preferably be efficient.
 * Note that it does not need to report anything except equal/unequal.
 * Note also that the length is exact, and the comparison should not
 * stop at embedded NULs!
 ^ static int casecmp(CONST chr *, CONST chr *, size_t);
 */
static	int			/* 0 for equal, nonzero for unequal */
casecmp_2 (
	register CONST TBufStringMarker*	xp,
	register CONST TBufStringMarker*	yp,
	register size_t						len)		/* exact length of comparison */
{
	TBufStringMarker	x	= *xp ;
	TBufStringMarker	y	= *yp ;

	while (len > 0) {
		if (Char_ToLower (TBufStringMarker_GetChar (&x)) != Char_ToUpper (TBufStringMarker_GetChar (&y)))
			return 1 ;
		TBufStringMarker_Forward (&x, 1) ;
		TBufStringMarker_Forward (&y, 1) ;
		len	-- ;
	}
	return	0 ;
}

#include "rege_dfa_2.c"


